@@ -37,6 +37,7 @@
 #include <stdlib.h>
 #include <stdio.h>
 #include <stdarg.h>
+#include <glib/gasyncqueue.h>
 #include <gtk/gtk.h>
 #include <libintl.h>
 #include <locale.h>
@@ -114,6 +115,8 @@ struct _AdvancedBacklightPluginPrivate
   GtkWidget *volume_slider;
   GtkWidget *eventbox;
   GtkWidget *rotation_button[4];
+  GtkWidget *brightness_level_label;
+  GtkWidget *volume_level_label;
 
   gboolean in_area;
   gboolean on_border;
@@ -124,9 +127,8 @@ struct _AdvancedBacklightPluginPrivate
 
   gboolean has_rotation;
   gboolean has_dsme;
+  gboolean all_done;
 
-  Display *dpy;
-  gint screen;
   GConfClient *gconf_client;
   void *dl_handle;
 
@@ -134,6 +136,7 @@ struct _AdvancedBacklightPluginPrivate
 
 
   gint is_n810;
+  gint is_770;
   struct timeval last_beep;
   gint brightness_jump;
   gint volume_jump;
@@ -141,6 +144,15 @@ struct _AdvancedBacklightPluginPrivate
   gboolean allow_zero;
 
   gint brightness_level;
+  gint volume_level;
+
+  GThread *brightness_thread;
+  GAsyncQueue *brightness_queue;
+  GString *brightness_level_string;
+  GString *brightness_level_dsme_string;
+  GThread *volume_thread;
+  GAsyncQueue *volume_queue;
+  GString *volume_level_string;
 
 };
 
@@ -165,8 +177,14 @@ static void advanced_backlight_plugin_finalize (GObject * object);
 static void show_menu (GtkButton * widget, AdvancedBacklightPlugin *);
 static void advanced_backlight_plugin_create_menu (AdvancedBacklightPlugin *);
 static void advanced_backlight_plugin_check_init_rotation (AdvancedBacklightPlugin *);
+static void advanced_backlight_plugin_init_threads (AdvancedBacklightPlugin *);
 static void advanced_backlight_plugin_init_dsme (AdvancedBacklightPlugin *);
-static void advanced_backlight_plugin_dsme_update_brightness (AdvancedBacklightPlugin *);
+static void advanced_backlight_plugin_dsme_update_brightness_value (AdvancedBacklightPluginPrivate *, gint);
+static void advanced_backlight_plugin_770_update_brightness_value (AdvancedBacklightPluginPrivate *, gint);
+static void advanced_backlight_plugin_update_volume_value (AdvancedBacklightPlugin *, gint);
+static gpointer advanced_backlight_plugin_dsme_update_brightness_thread (gpointer);
+static gpointer advanced_backlight_plugin_update_volume_thread (gpointer);
+/*static void advanced_backlight_plugin_dsme_update_brightness (AdvancedBacklightPlugin *);*/
 static void advanced_backlight_plugin_configure_applet (AdvancedBacklightPlugin * vsp);
 static void advanced_backlight_plugin_do_volume_beep (AdvancedBacklightPlugin * vsp);
 static void advanced_backlight_plugin_load_icons (AdvancedBacklightPlugin * vsp);
@@ -181,6 +199,7 @@ static void advanced_backlight_plugin_save_settings (AdvancedBacklightPlugin * v
 static void advanced_backlight_plugin_fill_theme_list (AdvancedBacklightPlugin * vsp, GtkWidget * combobox);
 static int get_light_sensor(AdvancedBacklightPlugin * vsp);
 static gboolean is_n810 ();
+static gboolean is_770();
 
 #ifdef XRANDR
 /* D-Pad mapping */
@@ -255,10 +274,12 @@ HD_DEFINE_PLUGIN (AdvancedBacklightPlugin, advanced_backlight_plugin, STATUSBAR_
 
   priv->gconf_client = gconf_client_get_default ();
 
+  advanced_backlight_plugin_init_threads (vsp);
   advanced_backlight_plugin_init_dsme (vsp);
   advanced_backlight_plugin_check_init_rotation (vsp);
   /* MERGED from ABL */
   priv->is_n810 = is_n810 ();
+  priv->is_770 = is_770 ();
   gettimeofday (&priv->last_beep, NULL);
   advanced_backlight_plugin_load_settings (vsp);
 
@@ -294,6 +315,20 @@ advanced_backlight_plugin_finalize (GObject * object)
 
   advanced_backlight_plugin_save_settings (vsp);
 
+  priv->all_done = TRUE;
+
+  g_string_free(priv->brightness_level_string, TRUE);
+  g_async_queue_push(priv->brightness_queue, GINT_TO_POINTER(priv->brightness_level));
+  g_thread_join(priv->brightness_thread);
+  g_async_queue_unref(priv->brightness_queue);
+  g_async_queue_unref(priv->brightness_queue);
+
+  g_string_free(priv->volume_level_string, TRUE);
+  g_async_queue_push(priv->volume_queue, vsp);
+  g_thread_join(priv->volume_thread);
+  g_async_queue_unref(priv->volume_queue);
+  g_async_queue_unref(priv->volume_queue);
+
   /* Destroy local widgets */
   if (priv->menu)
     {
@@ -303,6 +338,21 @@ advanced_backlight_plugin_finalize (GObject * object)
   G_OBJECT_CLASS (g_type_class_peek_parent (G_OBJECT_GET_CLASS (object)))->finalize (object);
 }
 
+static void
+advanced_backlight_plugin_init_threads (AdvancedBacklightPlugin * vsp)
+{
+  AdvancedBacklightPluginPrivate *priv = ADVANCED_BACKLIGHT_PLUGIN_GET_PRIVATE (vsp);
+
+  if (! g_thread_supported()) g_thread_init(NULL);
+
+  priv->all_done = FALSE;
+
+  priv->brightness_queue = g_async_queue_new();
+  priv->brightness_thread = g_thread_create(advanced_backlight_plugin_dsme_update_brightness_thread, vsp, TRUE, NULL);
+
+  priv->volume_queue = g_async_queue_new();
+  priv->volume_thread = g_thread_create(advanced_backlight_plugin_update_volume_thread, vsp, TRUE, NULL);
+}
 
 /*
     Dynamically load dsme library functions so it will build without it.
@@ -671,10 +721,48 @@ brightness_level_changed (GtkWidget * widget, AdvancedBacklightPlugin * vsp)
   g_debug ("%s\n", __PRETTY_FUNCTION__);
 
   priv->brightness_level = (gint) ((hildon_volumebar_get_level (HILDON_VOLUMEBAR (priv->brightness_slider)) * 127.0) / 100.0) + 1;
+  g_string_printf(priv->brightness_level_string, "%3d", priv->brightness_level - 1);
+  gtk_label_set_text(GTK_LABEL(priv->brightness_level_label), (priv->brightness_level_string)->str);
   g_debug ("value=%d\n", priv->brightness_level);
-
-  advanced_backlight_plugin_dsme_update_brightness (vsp);
   advanced_backlight_plugin_update_icons (vsp);
+
+  g_async_queue_push(priv->brightness_queue, GINT_TO_POINTER(priv->brightness_level));
+}
+
+static void
+advanced_backlight_plugin_update_volume_value (AdvancedBacklightPlugin * vsp, gint volume)
+{
+  AdvancedBacklightPluginPrivate *priv = ADVANCED_BACKLIGHT_PLUGIN_GET_PRIVATE (vsp);
+  gconf_client_set_int (priv->gconf_client, GCONF_MASTER_VOLUME, volume, NULL);
+  if (priv->volume_beep)
+    advanced_backlight_plugin_do_volume_beep (vsp);
+}
+
+static gpointer
+advanced_backlight_plugin_update_volume_thread (gpointer d)
+{
+  AdvancedBacklightPlugin *vsp = (AdvancedBacklightPlugin *)d;
+  AdvancedBacklightPluginPrivate *priv = ADVANCED_BACKLIGHT_PLUGIN_GET_PRIVATE (vsp);
+  gint i=0, j, value, prev_value=-101;
+  gpointer gip;
+
+  g_async_queue_ref(priv->volume_queue);
+  while ((gip = g_async_queue_pop(priv->volume_queue)) && ! priv->all_done) {
+    value = GPOINTER_TO_INT(gip);
+    j = 0;
+    g_async_queue_lock(priv->volume_queue);
+    while ((gip = g_async_queue_try_pop_unlocked(priv->volume_queue)) != NULL) {
+      value = GPOINTER_TO_INT(gip);
+      j++;
+    }
+    g_async_queue_unlock(priv->volume_queue);
+    if(0)g_debug("after queue pop j: %d %d\n", j, i);
+    if (prev_value != value) {
+      advanced_backlight_plugin_update_volume_value(vsp, value);
+      prev_value = value;
+    }
+  }
+  return g_thread_self();
 }
 
 /*
@@ -693,10 +781,11 @@ advanced_backlight_plugin_update_volume (AdvancedBacklightPlugin * vsp)
   volume = (gint) hildon_volumebar_get_level (HILDON_VOLUMEBAR (priv->volume_slider));
   if (mute)
     volume *= -1;
-  gconf_client_set_int (priv->gconf_client, GCONF_MASTER_VOLUME, volume, NULL);
+  g_string_printf(priv->volume_level_string, "%4d", volume);
+  gtk_label_set_text(GTK_LABEL(priv->volume_level_label), (priv->volume_level_string)->str);
+  priv->volume_level = volume;
   advanced_backlight_plugin_update_icons (vsp);
-  if (priv->volume_beep)
-    advanced_backlight_plugin_do_volume_beep (vsp);
+  g_async_queue_push(priv->volume_queue, GINT_TO_POINTER(volume));
 }
 
 static void
@@ -719,7 +808,7 @@ static void
 advanced_backlight_plugin_create_menu (AdvancedBacklightPlugin * vsp)
 {
   AdvancedBacklightPluginPrivate *priv = ADVANCED_BACKLIGHT_PLUGIN_GET_PRIVATE (vsp);
-  GtkWidget *hbox, *vbox, *label;
+  GtkWidget *hbox, *vbox, *label, *hbox2, *hbox3;
   PangoFontDescription *small_pfd;
   gint i;
 
@@ -789,19 +878,39 @@ advanced_backlight_plugin_create_menu (AdvancedBacklightPlugin * vsp)
   small_pfd = pango_font_description_copy (priv->menu->style->font_desc);
   pango_font_description_set_size (small_pfd, (pango_font_description_get_size (small_pfd) * 3) / 4);
 
+  hbox2 = gtk_hbox_new (TRUE, 0);
+
   label = gtk_label_new (_("Brightness"));
   gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.5);
   gtk_widget_modify_font (GTK_WIDGET (label), small_pfd);
-  gtk_box_pack_start (GTK_BOX (vbox), label, FALSE, FALSE, 0);
+  gtk_box_pack_start (GTK_BOX (hbox2), label, FALSE, FALSE, 0);
 
+  priv->brightness_level_string = g_string_new("   ");
+  g_string_printf(priv->brightness_level_string, "%3d", priv->brightness_level - 1);
+  priv->brightness_level_label = gtk_label_new ((priv->brightness_level_string)->str);
+  gtk_misc_set_alignment (GTK_MISC (priv->brightness_level_label), 0.0, 0.5);
+  gtk_widget_modify_font (GTK_WIDGET (priv->brightness_level_label), small_pfd);
+  gtk_box_pack_start (GTK_BOX (hbox2), priv->brightness_level_label, TRUE, FALSE, 0);
+
+  gtk_box_pack_start (GTK_BOX (vbox), hbox2, FALSE, FALSE, 0);
   gtk_box_pack_start (GTK_BOX (vbox), priv->brightness_slider, FALSE, FALSE, 0);
   gtk_box_pack_start (GTK_BOX (vbox), gtk_separator_menu_item_new (), FALSE, FALSE, 0);
 
+  hbox3 = gtk_hbox_new (TRUE, 0);
+
   label = gtk_label_new (_("Volume"));
   gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.5);
   gtk_widget_modify_font (GTK_WIDGET (label), small_pfd);
-  gtk_box_pack_start (GTK_BOX (vbox), label, FALSE, FALSE, 0);
+  gtk_box_pack_start (GTK_BOX (hbox3), label, FALSE, FALSE, 0);
 
+  priv->volume_level_string = g_string_new("   ");
+  g_string_printf(priv->volume_level_string, "%4d", 50);
+  priv->volume_level_label = gtk_label_new ((priv->volume_level_string)->str);
+  gtk_misc_set_alignment (GTK_MISC (priv->volume_level_label), 0.0, 0.5);
+  gtk_widget_modify_font (GTK_WIDGET (priv->volume_level_label), small_pfd);
+  gtk_box_pack_start (GTK_BOX (hbox3), priv->volume_level_label, TRUE, FALSE, 0);
+
+  gtk_box_pack_start (GTK_BOX (vbox), hbox3, FALSE, FALSE, 0);
   gtk_box_pack_start (GTK_BOX (vbox), priv->volume_slider, FALSE, FALSE, 0);
   gtk_box_pack_start (GTK_BOX (vbox), gtk_separator_menu_item_new (), FALSE, FALSE, 0);
 
@@ -1045,7 +1154,29 @@ advanced_backlight_plugin_set_rotation (AdvancedBacklightPlugin * vsp, gint rota
 }
 #endif /* XRANDR */
 
+static void
+advanced_backlight_plugin_dsme_update_brightness_value (AdvancedBacklightPluginPrivate *priv, gint value)
+{
+  struct dsme_message message;
+  gint conn;
 
+  if (!priv->has_dsme)
+    return;
+
+  value = (priv->brightness_level * 2);
+  if (value <= 2 && !priv->allow_zero)
+    value = 3;
+  if (value > 254)
+    value = 254;
+
+  conn = dsmesock_connect ();
+  message.type = 0x0c;
+  message.request = 0x289;
+  message.value = (void *) value;
+  dsmesock_send (conn, &message);
+  dsmesock_close (conn);
+}
+/*
 static void
 advanced_backlight_plugin_dsme_update_brightness (AdvancedBacklightPlugin * vsp)
 {
@@ -1061,6 +1192,8 @@ advanced_backlight_plugin_dsme_update_brightness (AdvancedBacklightPlugin * vsp)
   value = (priv->brightness_level * 2);
   if (value <= 2 && !priv->allow_zero)
     value = 3;
+  if (value > 254)
+    value = 254;
 
   conn = dsmesock_connect ();
   message.type = 0x0c;
@@ -1069,6 +1202,49 @@ advanced_backlight_plugin_dsme_update_brightness (AdvancedBacklightPlugin * vsp)
   dsmesock_send (conn, &message);
   dsmesock_close (conn);
 }
+*/
+static gpointer
+advanced_backlight_plugin_dsme_update_brightness_thread (gpointer d)
+{
+  AdvancedBacklightPlugin *vsp = (AdvancedBacklightPlugin *)d;
+  AdvancedBacklightPluginPrivate *priv = ADVANCED_BACKLIGHT_PLUGIN_GET_PRIVATE (vsp);
+  gint i=0, j, value, prev_value=-1;
+  gpointer gip;
+
+  g_async_queue_ref(priv->brightness_queue);
+  while ((gip = g_async_queue_pop(priv->brightness_queue)) && ! priv->all_done) {
+    value = GPOINTER_TO_INT(gip);
+    j = 0;
+    g_async_queue_lock(priv->brightness_queue);
+    while ((gip = g_async_queue_try_pop_unlocked(priv->brightness_queue)) != NULL) {
+      value = GPOINTER_TO_INT(gip);
+      j++;
+    }
+    g_async_queue_unlock(priv->brightness_queue);
+    if(0)g_debug("after queue pop j: %d %d\n", j, i);
+    if (prev_value != value) {
+      if (!is_770())
+        advanced_backlight_plugin_dsme_update_brightness_value(priv, value);
+      else 
+        advanced_backlight_plugin_770_update_brightness_value(priv, value);
+      prev_value = value;
+    }
+  }
+  return g_thread_self();
+}
+
+static void
+advanced_backlight_plugin_770_update_brightness_value (AdvancedBacklightPluginPrivate *priv, gint value)
+{
+  value = value * 256 + 128; // Math stolen directly from Fanoush's blset script.
+
+  FILE* myfile = fopen ("/sys/devices/platform/omapfb/panel/backlight_level", "w");
+
+  if (myfile != NULL)
+    fprintf(myfile, "%d\n", value);
+
+  fclose(myfile);
+}
 
 /*
  TODO: Find a way to fetch the brightness...
@@ -1103,6 +1279,21 @@ is_n810 ()
   return (version == '4');      /* 4 means n810, 3 means n800. */
 }
 
+static gboolean
+is_770 ()
+{
+  char version;
+
+  FILE *fp = fopen ("/proc/component_version", "r");
+  if (!fp)
+    return FALSE;
+  fseek (fp, sizeof (char) * 15, 0);    /* The important character is the 16th in the file. */
+  fread (&version, sizeof (char), 1, fp);
+  fclose (fp);
+
+  return (version == '1');      /* 4 means n810, 3 means n800. */
+}
+
 static void
 advanced_backlight_plugin_do_volume_beep (AdvancedBacklightPlugin * vsp)
 {
@@ -1131,7 +1322,8 @@ advanced_backlight_plugin_load_settings (AdvancedBacklightPlugin * vsp)
   if (conf_value)
     {
       priv->brightness_level = gconf_value_get_int (conf_value);
-      advanced_backlight_plugin_dsme_update_brightness (vsp);
+      /* advanced_backlight_plugin_dsme_update_brightness (vsp); */
+      g_async_queue_push(priv->brightness_queue, GINT_TO_POINTER(priv->brightness_level));
       gconf_value_free (conf_value);
     }
   else
@@ -1391,8 +1583,8 @@ advanced_backlight_plugin_update_icons (AdvancedBacklightPlugin * vsp)
 
   g_debug ("%s\n", __PRETTY_FUNCTION__);
 
-  volume = gconf_client_get_int (priv->gconf_client, GCONF_MASTER_VOLUME, NULL);
-  if (volume < 0)
+  volume = priv->volume_level;
+  if (volume < 0 || hildon_volumebar_get_mute (HILDON_VOLUMEBAR (priv->volume_slider)))
     volume_icon = NUM_VOLUME_ICONS - 1;
   else
     {
